@@ -172,8 +172,7 @@ namespace DDM {
             return;
 
         for (auto *auth : std::as_const(auths)) {
-            disconnect(auth, &Auth::userProcessFinished, nullptr, nullptr);
-            daemonApp->displayManager()->RemoveSession(auth->sessionId);
+            disconnect(auth, &Auth::sessionFinished, nullptr, nullptr);
             delete auth;
         }
         auths.clear();
@@ -300,6 +299,7 @@ namespace DDM {
         } else if (session.xdgSessionType() == QLatin1String("x11")) {
             auth->type = X11;
 
+            daemonApp->treelandConnector()->disconnect();
             m_treeland->stop();
             QThread::msleep(500); // give some time to treeland to stop properly
 
@@ -317,14 +317,13 @@ namespace DDM {
         } else {
             auth->type = Wayland;
 
+            daemonApp->treelandConnector()->disconnect();
             m_treeland->stop();
             QThread::msleep(500); // give some time to treeland to stop properly
         }
 
         // Open Logind session & Exec the desktop process
-        VirtualTerminal::setVtSignalHandler(nullptr, nullptr);
         int xdgSessionId = auth->openSession(session.exec(), env, cookie);
-        daemonApp->treelandConnector()->setSignalHandler();
 
         if (xdgSessionId <= 0) {
             qCritical() << "Failed to open logind session for user" << user;
@@ -338,8 +337,8 @@ namespace DDM {
             activateSession(auth->user, xdgSessionId);
         }
 
-        connect(auth, &Auth::userProcessFinished, this, [this, auth](int status) {
-            qWarning() << "Session for user" << auth->user << "finished with status" << status;
+        connect(auth, &Auth::sessionFinished, this, [this, auth]() {
+            qWarning() << "Session for user" << auth->user << "finished";
             auths.removeAll(auth);
             daemonApp->displayManager()->RemoveSession(auth->sessionId);
             delete auth;
@@ -353,33 +352,17 @@ namespace DDM {
     }
 
     void Display::logout([[maybe_unused]] QLocalSocket *socket, int id) {
-        for (Auth *auth : std::as_const(auths)) {
-            if (auth->xdgSessionId == id) {
-                disconnect(auth, &Auth::userProcessFinished, nullptr, nullptr);
-                auths.removeAll(auth);
-                daemonApp->displayManager()->RemoveSession(auth->sessionId);
-                delete auth;
-                break;
-            }
-        }
+        // Do not kill the session leader process before
+        // TerminateSession! Logind will only kill the session's
+        // cgroup (session_stop_scope) when the session is not in
+        // "stopping" state, killing the session leader beforehand
+        // will put the session in "stopping" early.
+
+        // https://github.com/systemd/systemd/blob/main/src/login/logind-session.c#L938
         OrgFreedesktopLogin1ManagerInterface manager(Logind::serviceName(),
                                                      Logind::managerPath(),
                                                      QDBusConnection::systemBus());
         manager.TerminateSession(QString::number(id));
-        QTimer::singleShot(2000, this, [this, id]() {
-            OrgFreedesktopLogin1ManagerInterface manager(Logind::serviceName(),
-                                                         Logind::managerPath(),
-                                                         QDBusConnection::systemBus());
-            // Some child processes may linger
-            // (e.g. gnome-keyring-daemon), kill them to avoid the
-            // whole session stuck in closing state. Send SIGKILL
-            // directly may make systemd-logind be left in a bad state
-            // and cannot handle future login, so send SIGTERM
-            // instead.
-
-            // TODO: Find out why
-            manager.KillSession(QString::number(id), "all", 15);
-        });
     }
 
     void Display::unlock(QLocalSocket *socket, const QString &user, const QString &password) {
